import {
  createArchive,
  createArchiveDb,
  createArchiveRepo,
  deleteArchive,
  addTarToArchive,
} from "../src/lib/archive";
import { createArchiveTar } from "./mocks/git";
import { repoPath } from "../src/lib/git_routines";
import db from "../src/lib/db";
const knexfile = require("../knexfile");
const fs = require("fs");
const rimraf = require("rimraf");
const path = require("path");

beforeAll(async () => {
  await db.migrate.latest();
});

afterAll(async () => {
  await db.destroy();
  try {
    fs.unlinkSync(knexfile.connection.filename);
  } catch (e) {
    console.log(e);
  }
  rimraf.sync(path.resolve("test", "reposTmp"));
});

beforeEach(async () => {
  rimraf.sync(path.resolve("test", "reposTmp"));
});

describe("archive", () => {
  const universe_name = "testuniv";
  const project_name = "testproj";
  const data = {
    universe_name,
    project_name,
    name: "testArc1",
    default_commit: "master",
  };
  test("createArchiveDb() returns id & inserts to repos table", async () => {
    const res = await createArchiveDb(data);
    expect(res).toBeGreaterThanOrEqual(1);
    expect((await db("repositories").where(data)).length).toBe(1);
  });
  test("createArchiveRepo() creates a valid repo & returns directory", async () => {
    const id = await createArchiveDb(data);
    const output = await createArchiveRepo(id);
    expect(output).toEqual(repoPath(id));
  });
  test("createArchive() creates a valid repo and inserts into repos table, and returns a directory", async () => {
    const { archive, archivePath } = await createArchive(data);
    expect(archive.id).toBeGreaterThanOrEqual(1);
    expect(archivePath).toEqual(repoPath(archive.id));
    expect(fs.existsSync(archivePath)).toBe(true);
  });
  test("deleteArchive() deletes repo folder & db entry", async () => {
    const { archive, archivePath } = await createArchive(data);
    await deleteArchive(archive.id);
    expect((await db("repositories").where({ id: archive.id })).length).toBe(0);
    expect(archivePath).toEqual(repoPath(archive.id));
    expect(fs.existsSync(archivePath)).toBe(false);
  });
  test("addTarToArchive() creates commit with files", async () => {
    const id = await createArchiveDb(data);
    await createArchiveRepo(id);
    const tarPath = createArchiveTar("foo", [
      {
        msg: "my commit",
        files: [
          { filepath: "foo", txt: "this is foo" },
          { filepath: "blah/foo2", txt: "this is blah/foo2" },
        ],
      },
    ]);
    const out = await addTarToArchive(tarPath, "foo", id);
    expect(out).toEqual(expect.stringMatching(/^[a-f0-9\s]+$/));
  });
});
