import {
  startSSHAgent,
  getAgent,
  addPrivateKeyToAgent,
  killSSHAgent,
} from "../src/lib/ssh";
import { readFileSync } from "fs";
import { newGit } from "../src/lib/git_routines";
import path from "path";
import fs from "fs";
import rimraf from "rimraf";

const testDir = __dirname;
const privKey = readFileSync(path.join(testDir, "corpus/ssh_key/key"), "utf8");
const invalidPrivKey = privKey.replace("a", "b");
const unusedPrivKey = readFileSync(
  path.join(testDir, "corpus/ssh_key/unused_key/unused"),
  "utf8"
);

// Open repo
const openRepoUrl = "https://github.com/rashkov-at-sauce/open_repo.git";

// SSH repo
const sshRepoUrl = "git@github.com:rashkov-at-sauce/private_repo.git";

// user/pass repo. Same as the SSH repo but with "https:"
const userPassRepoUrl = "https://github.com/rashkov-at-sauce/private_repo.git";
const user = "rashkov-at-sauce";
// This is a github "classic" token, accessible under Settings / Developer Settings / Tokens (Classic)
// Github now requires the use of tokens instead of your account password (which will not work)
// I've set this token to never expire, but if this account requires administration you can log in using:
// username: mike.rashkovsky@saucelabs.com and password: skyline-hurler-maker
// This account is only used for this test script and nothing else.
const pass = "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";

describe("SSH", () => {
  test("Start SSH agent", async () => {
    const { SSH_AUTH_SOCK, SSH_AGENT_PID } = await startSSHAgent();
    const pid = parseInt(SSH_AGENT_PID);
    expect(SSH_AUTH_SOCK).toBeDefined();
    expect(SSH_AUTH_SOCK).toContain("/tmp/ssh-");
    expect(pid).toBeDefined();
    expect(pid).toBeGreaterThan(0);
    expect(pid).not.toBeNaN();
  });
  test("getAgent() re-uses the same agent until killed", async () => {
    const { SSH_AUTH_SOCK, SSH_AGENT_PID } = await getAgent();
    const pid = parseInt(SSH_AGENT_PID);
    expect(SSH_AUTH_SOCK).toBeDefined();
    expect(SSH_AUTH_SOCK).toContain("/tmp/ssh-");
    expect(pid).toBeDefined();
    expect(pid).toBeGreaterThan(0);
    expect(pid).not.toBeNaN();

    const { SSH_AUTH_SOCK: sock2, SSH_AGENT_PID: pid2 } = await getAgent();
    expect(sock2).toBe(SSH_AUTH_SOCK);
    expect(pid2).toBe(SSH_AGENT_PID);

    killSSHAgent();
    const { SSH_AUTH_SOCK: sock3, SSH_AGENT_PID: pid3 } = await getAgent();
    expect(sock3).not.toBe(SSH_AUTH_SOCK);
    expect(pid3).not.toBe(SSH_AGENT_PID);
  });
  test("Add invalid private key to SSH agent", async () => {
    const privateKey = invalidPrivKey;

    // suppress console.error output
    const consoleError = console.error;
    console.error = () => {};

    const addKey = await addPrivateKeyToAgent(privateKey);

    // restore console.error
    console.error = consoleError;

    expect(addKey).toBe(false);
  });
  test("Add valid private key to SSH agent", async () => {
    const addKey = await addPrivateKeyToAgent(privKey);
    expect(addKey).toBe(true);
  });
});

describe("Clones and fetches an open / password-less repo", () => {
  test("clones and fetches a password-less repo", async () => {
    const url = openRepoUrl;
    const clonePath = path.resolve("test", "reposTmp", "open_repo_test");
    if (fs.existsSync(clonePath)) {
      rimraf.sync(clonePath);
    }
    const clonePromise = newGit.cloneOpenExn(url, clonePath);
    // expect clonePromise not to throw
    expect(clonePromise).resolves.not.toThrow();

    // check that it fetches successfully
    // first we finish cloning
    await clonePromise;
    // expect fetchPromise not to throw
    const fetchPromise = newGit.fetchOpenExn(clonePath);
    expect(fetchPromise).resolves.not.toThrow();
  });
  test("Fails to clone a non-existing repo", async () => {
    const url = openRepoUrl + "bad";
    const clonePath = path.resolve("test", "reposTmp", "open_repo_test");
    if (fs.existsSync(clonePath)) {
      rimraf.sync(clonePath);
    }
    const clonePromise = newGit.cloneOpenExn(url, clonePath);
    expect(clonePromise).rejects.toThrow();
  });
});

describe("Clones and fetches a private SSH-protected repo", () => {
  test("Successfully clones and fetches a private repo", async () => {
    const url = sshRepoUrl;
    const clonePath = path.resolve("test", "reposTmp", "ssh_repo_test");
    if (fs.existsSync(clonePath)) {
      rimraf.sync(clonePath);
    }
    const clonePromise = newGit.cloneSSHExn(url, clonePath, privKey);
    // expect clonePromise not to throw
    expect(clonePromise).resolves.not.toThrow();

    // check that it fetches successfully
    // first we finish cloning
    await clonePromise;
    // expect fetchPromise not to throw
    const fetchPromise = newGit.fetchSSHExn(clonePath, privKey);
    expect(fetchPromise).resolves.not.toThrow();
  });
  test("Fails to clone a non-existing repo", async () => {
    const url = openRepoUrl + "bad";
    const clonePath = path.resolve("test", "reposTmp", "ssh_repo_test");
    if (fs.existsSync(clonePath)) {
      rimraf.sync(clonePath);
    }
    const clonePromise = newGit.cloneSSHExn(url, clonePath, privKey);
    expect(clonePromise).rejects.toThrow();
  });
  test("Fails to clone a repo with invalid credentials", async () => {
    // remove valid ssh key from agent
    killSSHAgent();
    const url = sshRepoUrl;
    const clonePath = path.resolve("test", "reposTmp", "ssh_repo_test");
    if (fs.existsSync(clonePath)) {
      rimraf.sync(clonePath);
    }
    const clonePromise = newGit.cloneSSHExn(url, clonePath, unusedPrivKey);
    expect(clonePromise).rejects.toThrow();
  });
});

describe("Clones and fetches a username/password protected repo", () => {
  test("Successfully clones and fetches a private repo", async () => {
    const clonePath = path.resolve("test", "reposTmp", "user_pass_test");
    if (fs.existsSync(clonePath)) {
      rimraf.sync(clonePath);
    }
    const clonePromise = newGit.cloneUserPassExn(
      userPassRepoUrl,
      clonePath,
      user,
      pass
    );
    // expect clonePromise not to throw
    expect(clonePromise).resolves.not.toThrow();

    // check that it fetches successfully
    // first we finish cloning
    await clonePromise;
    const fetchPromise = newGit.fetchUserPassExn(clonePath);
    // expect fetchPromise not to throw
    expect(fetchPromise).resolves.not.toThrow();
  });
  test("Fails to clone a non-existing repo", async () => {
    const url = userPassRepoUrl + "bad";
    const clonePath = path.resolve("test", "reposTmp", "user_pass_test");
    if (fs.existsSync(clonePath)) {
      rimraf.sync(clonePath);
    }
    const clonePromise = newGit.cloneUserPassExn(url, clonePath, user, pass);
    // expect clonePromise to throw
    expect(clonePromise).rejects.toThrow();
  });
  test("Fails to clone a repo with invalid credentials", async () => {
    let badUser = user + "bad";
    let badPass = pass + "bad";
    const clonePath = path.resolve("test", "reposTmp", "user_pass_test");
    if (fs.existsSync(clonePath)) {
      rimraf.sync(clonePath);
    }
    const clonePromise = newGit.cloneUserPassExn(
      userPassRepoUrl,
      clonePath,
      badUser,
      badPass
    );
    // expect clonePromise to throw
    expect(clonePromise).rejects.toThrow();
  });
});

describe("Git remote", () => {
  const url = openRepoUrl;
  const clonePath = path.resolve("test", "reposTmp", "open_repo_test");

  beforeEach(async () => {
    // clone an open repo
    if (fs.existsSync(clonePath)) {
      rimraf.sync(clonePath);
    }
    await newGit.cloneOpenExn(url, clonePath);
  });

  test("Lists remotes", async () => {
    const remotes = await newGit.listRemotes(clonePath);
    expect(remotes.length).toBe(2); // "push" and "fetch" remote
    const origin = remotes.find((r) => r[0] == "origin");
    expect(origin).toBeDefined();
    expect(origin[0]).toBe("origin");
    expect(origin[1]).toMatch("https");
  });
  test("Removes remote", async () => {
    const remotes = await newGit.listRemotes(clonePath);
    expect(remotes.length).toBe(2); // "push" and "fetch" remote
    const origin = remotes.find((r) => r[0] == "origin");
    expect(origin).toBeDefined();
    expect(origin[0]).toBe("origin");
    expect(origin[1]).toMatch("https");

    await newGit.removeRemote(clonePath, "origin");
    const remotesAfter = await newGit.listRemotes(clonePath);
    expect(remotesAfter.length).toBe(0);
    const originAfter = remotesAfter.find((r) => r[0] == "origin");
    expect(originAfter).toBeUndefined();
  });
  test("Quietly attempts to remove non-existing remote", async () => {
    const remotes = await newGit.listRemotes(clonePath);
    expect(remotes.length).toBe(2); // "push" and "fetch" remote
    await newGit.removeRemote(clonePath, "origin2");
    const remotesAfter = await newGit.listRemotes(clonePath);
    expect(remotesAfter.length).toBe(2);
  });
  test("Adds remote", async () => {
    const remotes = await newGit.listRemotes(clonePath);
    expect(remotes.length).toBe(2); // "push" and "fetch" remote

    const newRemoteName = "new-origin";
    const newRemoteUrl = openRepoUrl + "new";
    await newGit.addRemoteExn(clonePath, newRemoteName, newRemoteUrl);
    const remotesAfter = await newGit.listRemotes(clonePath);

    expect(remotesAfter.length).toBe(4); // "push" and "fetch" remote for new remote
    const originAfter = remotesAfter.find((r) => r[0] == newRemoteName);
    expect(originAfter).toBeDefined();
    expect(originAfter[0]).toBe(newRemoteName);
    expect(originAfter[1]).toEqual(newRemoteUrl);
  });
  test("Throws when adding remote twice", async () => {
    const remotes = await newGit.listRemotes(clonePath);
    expect(remotes.length).toBe(2); // "push" and "fetch" remote

    const newRemoteName = "new-origin";
    const newRemoteUrl = openRepoUrl + "new";
    await newGit.addRemoteExn(clonePath, newRemoteName, newRemoteUrl);
    const secondAdditionPromise = newGit.addRemoteExn(
      clonePath,
      newRemoteName,
      newRemoteUrl
    );
    expect(secondAdditionPromise).rejects.toThrow();
  });
  test("addOrUpdateRemote adds a remote twice without throwing", async () => {
    const remotes = await newGit.listRemotes(clonePath);
    expect(remotes.length).toBe(2); // "push" and "fetch" remote

    const newRemoteName = "new-origin";
    const newRemoteUrl = openRepoUrl + "new";
    const newRemoteUrl2 = openRepoUrl + "new";
    await newGit.addOrUpdateRemote(clonePath, newRemoteName, newRemoteUrl);
    await newGit.addOrUpdateRemote(clonePath, newRemoteName, newRemoteUrl2);
    const remotesAfter = await newGit.listRemotes(clonePath);

    expect(remotesAfter.length).toBe(4); // "push" and "fetch" remote for new remote
    const newOriginAfter = remotesAfter.find((r) => r[0] == newRemoteName);
    expect(newOriginAfter).toBeDefined();
    expect(newOriginAfter[0]).toBe(newRemoteName);
    expect(newOriginAfter[1]).toEqual(newRemoteUrl2); // should be updated
  });
});

describe("Git rev-parse", () => {
  const url = openRepoUrl;
  const clonePath = path.resolve("test", "reposTmp", "open_repo_test");

  // These are all the same commit
  const shortSha = "f95e9b";
  const longSha = "f95e9becbbbb7e14a787de4089e81adf4b5bc7ce";
  const tag = "first_commit";

  const invalidSha = "invalid_sha";

  beforeEach(async () => {
    // clone an open repo
    if (fs.existsSync(clonePath)) {
      rimraf.sync(clonePath);
    }
    await newGit.cloneOpenExn(url, clonePath);
  });

  test("revParseExn() converts short SHA to long SHA", async () => {
    const shortShaHash = await newGit.revParseExn(clonePath, shortSha);
    const longShaHash = await newGit.revParseExn(clonePath, longSha);
    expect(shortShaHash).toBe(longShaHash);
  });
  test("revParseExn() Converts a tag to a SHA", async () => {
    const tagHash = await newGit.revParseExn(clonePath, tag);
    const longShaHash = await newGit.revParseExn(clonePath, longSha);
    expect(tagHash).toBe(longShaHash);
  });
  test("revParseExn() Throws when given an invalid SHA", async () => {
    const invalidShaPromise = newGit.revParseExn(clonePath, invalidSha);
    expect(invalidShaPromise).rejects.toThrow();
  });
  test("revParse() returns NULL when given an invalid SHA", async () => {
    const invalidShaPromise = newGit.revParse(clonePath, invalidSha);
    expect(invalidShaPromise).resolves.toBeNull();
  });
  test("getCommit() calls revParse()", async () => {
    // spy on revParse
    const revParseSpy = jest.spyOn(newGit, "revParse");
    const shortShaHash = await newGit.getCommit(clonePath, shortSha);
    const longShaHash = await newGit.getCommit(clonePath, longSha);
    expect(shortShaHash).toBe(longShaHash);
    expect(revParseSpy).toHaveBeenCalledTimes(2);
  });
});
