import { getMimeType, decompress } from "../src/lib/util";
import {
  gitInit,
  appendText,
  writeText,
  gitAdd,
  gitCommitAll,
  gitShow,
  gitListDeletedUnstaged,
  createTar,
} from "./test_utils";
import { rmAllFromGit } from "../src/lib/archive";
import { Repository } from "nodegit";
import * as fs from "fs";
import { mktempFile, mktempDir } from "./test_utils";
import * as path from "path";
const { resolve } = path;

describe("test utils", () => {
  test("gitInit() returns valid repo path", async () => {
    const pathToRepo = gitInit();
    await Repository.open(pathToRepo);
  });
  test("appendText() creates a file (if it doesn't already exist)", async () => {
    const pathToRepo = gitInit();
    const path = resolve(pathToRepo, "foo");
    expect(fs.existsSync(path)).toBe(false);
    appendText(path, "hello");
    expect(fs.existsSync(path)).toBe(true);
  });
  test("appendText() creates path components (if they doesn't already exist)", async () => {
    const pathToRepo = gitInit();
    const path = resolve(pathToRepo, "blah/hey/foo");
    expect(fs.existsSync(path)).toBe(false);
    appendText(path, "hello");
    expect(fs.existsSync(path)).toBe(true);
  });
  test("appendText() appends text", async () => {
    const pathToRepo = gitInit();
    const path = resolve(pathToRepo, "foo");
    appendText(path, "hello");
    expect(fs.readFileSync(path).toString()).toMatch(/hello/);
    appendText(path, "world");
    expect(fs.readFileSync(path).toString()).toMatch(/helloworld/);
  });
  test("writeText() overwrites text", async () => {
    const pathToRepo = gitInit();
    const path = resolve(pathToRepo, "foo");
    writeText(path, "hello");
    expect(fs.readFileSync(path).toString()).toMatch(/^hello$/);
    writeText(path, "world");
    expect(fs.readFileSync(path).toString()).toMatch(/^world$/);
  });
  test("gitCommitAll() commits tracked & untracked files", async () => {
    const repoPath = gitInit();

    // tracked file
    const filename = "foo";
    const filepath = resolve(repoPath, filename);
    appendText(filepath, "this is foo");
    gitAdd(repoPath, filepath);

    // untracked file
    const filename2 = "bar";
    const filepath2 = resolve(repoPath, filename2);
    appendText(filepath2, "this is bar");

    gitCommitAll(repoPath, "my commit");

    expect(gitShow(repoPath, filename, "HEAD")).toBe("this is foo");
    expect(gitShow(repoPath, filename2, "HEAD")).toBe("this is bar");
  });
  test("rmAllFromGit() deletes all tracked files", async () => {
    const repoPath = gitInit();

    // tracked file
    const filename = "foo";
    const filepath = resolve(repoPath, filename);
    appendText(filepath, "this is foo");
    gitAdd(repoPath, filepath);

    // tracked file
    const filename2 = "blah/foo2";
    const filepath2 = resolve(repoPath, filename2);
    appendText(filepath2, "this is foo");
    gitAdd(repoPath, filepath2);

    gitCommitAll(repoPath, "my commit");

    await rmAllFromGit(repoPath);
    const lines = gitListDeletedUnstaged(repoPath).trim().split("\n");
    expect(lines.sort()).toEqual(["D\tfoo", "D\tblah/foo2"].sort());
  });
  test("getMimeType() returns a valid mime type string", async () => {
    const tmpFile = mktempFile();
    expect(getMimeType(tmpFile)).toBe("inode/x-empty; charset=binary");
  });
});

describe("general utils -- src/lib/util", () => {
  test("mktempFile() returns a filepath", async () => {
    const tmpFile = mktempFile();
    expect(typeof tmpFile).toBe("string");
  });
  describe("decompress()", () => {
    test("decompress works with tar", async () => {
      const filepath = createTar();
      const outdir = mktempDir();
      decompress(filepath, outdir, undefined, 1);
      expect(fs.existsSync(path.resolve(outdir, "okay"))).toBe(true);
      expect(fs.readFileSync(path.resolve(outdir, "okay")).toString()).toBe(
        "hello"
      );
    });
    test("decompress works with bz2", async () => {
      const filepath = createTar("bz2");
      const outdir = mktempDir();
      decompress(filepath, outdir, undefined, 1);
      expect(fs.existsSync(path.resolve(outdir, "okay"))).toBe(true);
      expect(fs.readFileSync(path.resolve(outdir, "okay")).toString()).toBe(
        "hello"
      );
    });
    test("decompress works with gz", async () => {
      const filepath = createTar("gz");
      const outdir = mktempDir();
      decompress(filepath, outdir, undefined, 1);
      expect(fs.existsSync(path.resolve(outdir, "okay"))).toBe(true);
      expect(fs.readFileSync(path.resolve(outdir, "okay")).toString()).toBe(
        "hello"
      );
    });
    test("decompress works with xz", async () => {
      const filepath = createTar("xz");
      const outdir = mktempDir();
      decompress(filepath, outdir, undefined, 1);
      expect(fs.existsSync(path.resolve(outdir, "okay"))).toBe(true);
      expect(fs.readFileSync(path.resolve(outdir, "okay")).toString()).toBe(
        "hello"
      );
    });
    test("decompress works with zip", async () => {
      const filepath = createTar("zip");
      const outdir = mktempDir();
      decompress(filepath, outdir, undefined, 1);
      expect(fs.existsSync(path.resolve(outdir, "okay"))).toBe(true);
      expect(fs.readFileSync(path.resolve(outdir, "okay")).toString()).toBe(
        "hello"
      );
    });
  });
});

describe("test mocks", () => {
  test("createTar creates a tar file", () => {
    const tarPath = createTar();
    expect(getMimeType(tarPath)).toMatch(/application\/x-tar/);
  });
  test("createTar creates a bz2 file", () => {
    const tarPath = createTar("bz2");
    expect(getMimeType(tarPath)).toMatch(/application\/x-bzip2/);
  });
  test("createTar creates a gz file", () => {
    const tarPath = createTar("gz");
    expect(getMimeType(tarPath)).toMatch(/application\/gzip/);
  });
  test("createTar creates a xz file", () => {
    const tarPath = createTar("xz");
    expect(getMimeType(tarPath)).toMatch(/application\/x-xz/);
  });
  test("createTar creates a zip file", () => {
    const tarPath = createTar("zip");
    expect(getMimeType(tarPath)).toMatch(/application\/zip/);
  });
});
